/******************************************************************************
 *                  QLOGIC LINUX SOFTWARE
 *
 * QLogic ISP2x00 device driver for Linux 2.2.x and 2.4.x
 * Copyright (C) 2002 Qlogic Corporation
 * (www.qlogic.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/

/****************************************************************************
              Please see revision.notes for revision history.
*****************************************************************************/


/*
* qla2x00_alloc_phys  (64-bit addressing)
*      Function used to allocate physical memory
*      and zero it.
*
* Input:
*      size     = size in bytes.
*      flag     = sleep flag.
*      phy_addr = physical address pointer.
*
* Returns:
*      not NULL = success
*/
STATIC void * qla2x00_alloc_phys(int siz, u_long *phy_addr)
{
    void        *mem = NULL;

    mem = KMALLOC(siz);
    if( mem != NULL ) {
        *phy_addr = VIRT_TO_BUS(mem);
        BZERO((caddr_t)mem, siz);
    }

    return(mem);
}


/************************************************************************/
/* qla2x00_ip_inquiry                                                   */
/*    This routine is called by the IP driver to find adapters that     */
/*    support IP and to get adapter parameters from the SCSI driver.    */
/*                                                                      */
/* Arguments:                                                           */
/*    wAdapterNumber - adapter number to check.                         */
/*    pInquiryData - pointer to BD_INQUIRY_DATA structure.              */
/*                                                                      */
/* Return Value:                                                        */
/*    Returns TRUE if specified adapter supports IP.                    */
/*    Returns FALSE if no IP support.                                   */
/************************************************************************/

int qla2x00_ip_inquiry(uint16_t wAdapterNumber,
                       BD_INQUIRY_DATA *pInquiryData)
{
    int status = FALSE;
    scsi_qla_host_t *ha;
    int i;

    /* Verify structure size and version */
    if (pInquiryData->wLength  == BDI_LENGTH &&
        pInquiryData->wVersion == BDI_VERSION)
    {
        /* Find the specified host adapter */
        for (ha = qla2x00_hostlist;
             ha != NULL && ha->host_no != wAdapterNumber;
             ha = ha->next);

        if (ha)
        {
            /* Found specified adapter, check for IP support */
            if (ha->device_id != QLA2100_DEVICE_ID && ha->flags.online)
            {
                DEBUG(printk("qla2x00_ip_inquiry: found adapter %d\n\r", wAdapterNumber);)

                /* Return inquiry data to backdoor IP driver */
                pInquiryData->lOptions                = BDI_IP_SUPPORT;
                pInquiryData->pScsiDriverHa           = ha;                
                pInquiryData->pReceiveBufferQueue     = ha->pIpBufferQueue;    
                pInquiryData->wReceiveBufferQueueSize = IP_BUFFER_QUEUE_DEPTH;
                for (i = 0; i < 8; i++)
                    pInquiryData->acPortName[i]       = ha->acPortName[i];          
                status = TRUE;
            }
        }
    }
    else
    {
        DEBUG(printk("qla2x00_ip_inquiry: incompatable structure\n\r");)
    }
    return(status);
} /* qla2x00_ip_inquiry */


/************************************************************************/
/* qla2x00_ip_enable                                                    */
/*    This routine is called by the IP driver to enable IP connection   */
/*    to the scsi driver and to pass in IP driver parameters.           */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - host adapter block pointer.                                  */
/*    pEnableData - pointer to BD_ENABLE_DATA structure.                */
/*                                                                      */
/* Return Value:                                                        */
/*    Returns TRUE if IP connection successfully enabled.               */
/*    Returns FALSE if IP not enabled.                                  */
/************************************************************************/

int qla2x00_ip_enable(scsi_qla_host_t *ha,
                      BD_ENABLE_DATA *pEnableData)
{
    int status = FALSE;

    DEBUG(printk("qla2x00_ip_enable: enable adapter %d\n\r", (int)ha->host_no);)

    /* Verify structure size and version and adapter online */
    if (pEnableData->wLength  == BDE_LENGTH  &&
        pEnableData->wVersion == BDE_VERSION &&
        ha->flags.online)
    {
        /* Save parameters from IP driver */
        ha->wReceiveBufferCount          = pEnableData->wReceiveBufferCount;
        ha->wHeaderSize                  = pEnableData->wHeaderSize;
        ha->lReceiveBufferSize           = pEnableData->lReceiveBufferSize;
        ha->lMaximumTransferSize         = pEnableData->lMaximumTransferSize;
        ha->pReturnReceivePacketsRoutine = (PRECEIVE_PACKETS_CALLBACK)pEnableData->pReturnReceivePacketsRoutine;
        ha->pReturnReceivePacketsContext = pEnableData->pReturnReceivePacketsContext;
        ha->pReceiveBufferCBs            = (PBUFFER_CB)pEnableData->pReceiveBufferCBs;
        ha->pSendCompletionRoutine       = (PSEND_COMPLETION_CALLBACK)pEnableData->pSendCompletionRoutine;
        if (pEnableData->lOptions & BDE_NOTIFY_ROUTINE)
        {
            ha->pNotifyRoutine           = (PNOTIFY_CALLBACK)pEnableData->pNotifyRoutine;
            ha->pNotifyContext           = pEnableData->pNotifyContext;
        }

        /* Enable RISC IP support */
        if (!(status = qla2x00_ip_initialize(ha)))
        {
            DEBUG(printk("qla2x00_ip_enable: IP init failed\n\r");)
            ha->pNotifyRoutine = NULL;
        }
    }
    else
    {
        DEBUG(printk("qla2x00_ip_enable: incompatable structure or offline\n\r");)
    }

    return(status);
} /* qla2x00_ip_enable */


/************************************************************************/
/* qla2x00_ip_disable                                                   */
/*    This routine is called by the IP driver to disable IP connection. */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - host adapter block pointer.                                  */
/*                                                                      */
/* Return Value:                                                        */
/*    none                                                              */
/************************************************************************/

void qla2x00_ip_disable(scsi_qla_host_t *ha)
{
	mbx_cmd_t       mc;
	mbx_cmd_t       *mcp = &mc;

    DEBUG(printk("qla2x00_ip_disable: disable adapter %d\n\r", (int)ha->host_no);)

    /* Disable IP support */
    ha->flags.enable_ip     = FALSE;

    mcp->mb[0] = MBC_DISABLE_IP;
	mcp->out_mb = MBX_0;
	mcp->in_mb = MBX_0;
	mcp->tov = 30;
	mcp->flags = 0;

    if (qla2x00_mailbox_command(ha, mcp) == QL_STATUS_SUCCESS)
    {
        /* IP disabled successful */
        DEBUG(printk("qla2x00_ip_disable: successful\n\r");)
    }
    else
    {
        DEBUG(printk("qla2x00_ip_disable: MBC_DISABLE_IP failed\n\r");)
    }

    /* Reset IP parameters */
    ha->wIpBufferIn         = 0;
    ha->wIpBufferOut        = 0;
    ha->pNotifyRoutine      = NULL;

} /* qla2x00_ip_disable */


/************************************************************************/
/* qla2x00_ip_initialize                                                */
/*    This function is called to initialize the RISC IP support.        */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*                                                                      */
/* Return Value:                                                        */
/*    Returns TRUE if initialization is complete                        */
/*    Returns FALSE if error                                            */
/************************************************************************/

static int qla2x00_ip_initialize(scsi_qla_host_t *ha)
{
    int             status = FALSE;
    int             i;
    IP_INIT_BLOCK   *pIpInitBlock;
    dma_addr_t  phys_address = 0;
	mbx_cmd_t       mc;
	mbx_cmd_t       *mcp = &mc;
    device_reg_t    *reg;

    DEBUG(printk("qla2x00_ip_initialize: enter\n\r");)

    /* Initialize IP data in ha */
    ha->pIpDeviceTop       = NULL;
    ha->pIpDeviceBottom    = NULL;

    ha->pIpDeviceFree      = &ha->asIpDeviceDatabase[0];
    for (i = 0; i < QLLAN_MAX_IP_DEVICES; i++)
    {
        ha->asIpDeviceDatabase[i].wDeviceIndex = i;
        ha->asIpDeviceDatabase[i].pNextIpDevice = &ha->asIpDeviceDatabase[i+1];
    }
    ha->asIpDeviceDatabase[QLLAN_MAX_IP_DEVICES-1].pNextIpDevice = NULL;

    /* Reset/pack buffers owned by RISC in receive buffer ring */
    if (ha->wIpBufferIn != ha->wIpBufferOut)
    {
        PBUFFER_CB       pBufferCB = (PBUFFER_CB)ha->pReceiveBufferCBs;
        uint16_t         wIpBufferOut = ha->wIpBufferOut;
        PIP_BUFFER_ENTRY pIpBuffer;

        /* Must locate all RISC owned buffers and pack them in the buffer ring */
        /* between IpBufferOut and IpBufferIN */
        for (i = 0; i < ha->wReceiveBufferCount; i++, pBufferCB++)
        {
            if (pBufferCB->lFlags & BCB_FLAGS_RISC_OWNS_BUFFER)
            {
                /* Set RISC owned buffer into receive buffer ring */
                pIpBuffer = &ha->asIpBuffers[wIpBufferOut];
                pIpBuffer->wIpBufferHandle = pBufferCB->wBufferHandle;
                pIpBuffer->lIpBufferAddrLow = pBufferCB->physicalAddressLow;
                pIpBuffer->lIpBufferAddrHigh = pBufferCB->physicalAddressHigh;

                if (wIpBufferOut < IP_BUFFER_QUEUE_DEPTH - 1)
                    wIpBufferOut++;
                else
                    wIpBufferOut = 0;
            }
        }

        /* Verify correct number of RISC owned buffers were found */
        if (wIpBufferOut != ha->wIpBufferIn)
        {
            /* Incorrect number of RISC owned buffers?? */
            DEBUG(printk("qla2x00_ip_initialize: incorrect number of RISC owned buffers, disable IP\n\r");)
            ha->flags.enable_ip = FALSE;
            return(FALSE);
        }
    }

    /* Init RISC buffer pointer */
    reg = ha->iobase;
    WRT_REG_WORD(&reg->mailbox8, ha->wIpBufferIn);

    /* Setup IP initialization control block */
    pIpInitBlock  = pci_alloc_consistent(ha->pdev, 
			sizeof(IP_INIT_BLOCK), 
			&phys_address);
    if (pIpInitBlock  != NULL)
    {
        BZERO((caddr_t)pIpInitBlock, sizeof(IP_INIT_BLOCK));
        pIpInitBlock->cVersion              = IPIB_VERSION;
        pIpInitBlock->wIpFirmwareOptions    = IPIB_OPTION_NO_BROADCAST_FASTPOST |
                                              IPIB_OPTION_64BIT_ADDRESSING;
        pIpInitBlock->wIpHeaderSize         = ha->wHeaderSize;
        pIpInitBlock->wIpMTUSize            = (uint16_t)ha->lMaximumTransferSize;
        pIpInitBlock->wIpBufferSize         = (uint16_t)ha->lReceiveBufferSize;
        pIpInitBlock->wIpQueueSize          = IP_BUFFER_QUEUE_DEPTH;
        pIpInitBlock->wIpLowWaterMark       = IPIB_LOW_WATER_MARK;
        pIpInitBlock->wIpQueueAddr[0]       = ha->ppIpBufferQueueLow & 0xffff;
        pIpInitBlock->wIpQueueAddr[1]       = ha->ppIpBufferQueueLow >> 16;
        pIpInitBlock->wIpQueueAddr[2]       = ha->ppIpBufferQueueHigh & 0xffff;
        pIpInitBlock->wIpQueueAddr[3]       = ha->ppIpBufferQueueHigh >> 16;
        pIpInitBlock->wIpQueueInPointer     = ha->wIpBufferOut;
        pIpInitBlock->wIpFastPostCount      = IPIB_FAST_POST_COUNT;
        pIpInitBlock->wIpBufferContainerCount = IPIB_BUFFER_CONTAINER_COUNT;
        pIpInitBlock->wIpIocbResourceAllocation = IPIB_IOCB_RESERVE_COUNT;

        /* Issue mailbox command to initialize IP firmware */
        mcp->mb[0] = MBC_INITIALIZE_IP;
        mcp->mb[2] = MSW(phys_address);
        mcp->mb[3] = LSW(phys_address);
        mcp->mb[6] = QL21_64BITS_4THWD(phys_address);
        mcp->mb[7] = QL21_64BITS_3RDWD(phys_address);
	    mcp->out_mb = MBX_7|MBX_6|MBX_3|MBX_2|MBX_0;
	    mcp->in_mb = MBX_0;
	    mcp->tov = 30;
    	mcp->buf_size = sizeof(IP_INIT_BLOCK);
	    mcp->flags = MBX_DMA_OUT;

        status = qla2x00_mailbox_command(ha, mcp);
        if(status == QL_STATUS_SUCCESS)
        {
            /* IP initialization successful */
            DEBUG(printk("qla2x00_ip_initialize: successful\n\r");)

            ha->flags.enable_ip = TRUE;

            /* Force database update */
            set_bit(LOOP_RESYNC_NEEDED, &ha->dpc_flags);
            set_bit(LOCAL_LOOP_UPDATE, &ha->dpc_flags);
            
            /* qla2x00_loop_resync(ha); */
    		if (ha->dpc_wait && !ha->dpc_active) {
        		up(ha->dpc_wait);
            }
            status = TRUE;
        }
        else
        {
            DEBUG(printk("qla2x00_ip_initialize: MBC_INITIALIZE_IP failed %x MB0 %x\n\r",
                  status, mcp->mb[0]);)
            status = FALSE;
        }
	pci_free_consistent(ha->pdev, sizeof(IP_INIT_BLOCK), 
			pIpInitBlock, phys_address);
        
    }
    else
    {
        DEBUG(printk("qla2x00_ip_initialize: memory allocation error\n\r");)
    }
    return(status);
} /* qla2200_ip_initialize */


/************************************************************************/
/* qla2x00_add_buffers                                                  */
/*    This routine is called directly by network IP backdoor driver     */
/*    to pass new buffers in the receive buffer queue to the ISP.       */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    wBufferCount - number of buffers added                            */
/*                                                                      */
/* Return Value:                                                        */
/*    None                                                              */
/************************************************************************/

void qla2x00_add_buffers(scsi_qla_host_t *ha,
                         uint16_t wBufferCount)
{
    int             i;
    uint16_t        wIpBufferIn = ha->wIpBufferIn;
    device_reg_t    *reg;
    PIP_BUFFER_ENTRY pIpBufferQueue = ha->asIpBuffers;
    PBUFFER_CB      pReceiveBufferCBs = (PBUFFER_CB)ha->pReceiveBufferCBs;
    uint16_t        wBufferHandle;

    /* Set RISC owns buffer flags */
    for (i = 0; i < wBufferCount; i++)
    {
        wBufferHandle = pIpBufferQueue[wIpBufferIn].wIpBufferHandle;
        pReceiveBufferCBs[wBufferHandle].lFlags |= BCB_FLAGS_RISC_OWNS_BUFFER;

        if (wIpBufferIn < IP_BUFFER_QUEUE_DEPTH - 1)
            wIpBufferIn++;
        else
            wIpBufferIn = 0;
    }

    /* Update RISC buffer pointer */
    reg = ha->iobase;
    WRT_REG_WORD(&reg->mailbox8, wIpBufferIn);
    ha->wIpBufferIn = wIpBufferIn;
} /* qla2x00_add_buffers */


/************************************************************************/
/* qla2x00_add_new_ip_device                                            */
/*    This routine reserves an IP block, saves IP data in block, and    */
/*    does fabric login if public device.                               */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    wLoopID   - loop ID if private loop, else PUBLIC_LOOP_DEVICE      */
/*    pPortId   - pointer to port ID if public device                   */
/*    pPortName - pointer to world wide port name                       */
/*    bForceAdd - if TRUE and resource error, logout other device       */
/*                                                                      */
/* Return Value:                                                        */
/*    QL_STATUS_SUCCESS         if no error                             */
/*    QL_STATUS_RESOURCE_ERROR  if out of loop IDs or IP device blocks  */
/*    QL_STATUS_FATAL_ERROR     if fatal error                          */
/*    QL_STATUS_ERROR           if any other error                      */
/************************************************************************/

static int qla2x00_add_new_ip_device(scsi_qla_host_t *ha,
                                     uint16_t wLoopId,
                                     uint8_t *pPortId,
                                     uint8_t *pPortName,
                                     int bForceAdd)
{
    IP_DEVICE_BLOCK      *pIpDevice;
    int                  status;

    /* Get free IP device block */
    if ((status = qla2x00_reserve_ip_block(ha, &pIpDevice)) == QL_STATUS_RESOURCE_ERROR)
    {
        if (bForceAdd)
        { 
            /* Out of IP blocks, bump public device at bottom of list */
            DEBUG(printk("qla2x00_add_new_ip_device: bump device from IP list\n\r");)

            for (pIpDevice = ha->pIpDeviceBottom; pIpDevice != NULL;
                 pIpDevice = pIpDevice->pLastIpDevice)
            {
                if (pIpDevice->lDeviceFlags & IP_DEV_FLAG_PUBLIC_DEVICE)
                {
                    /* Do fabric logout and free loop ID */
                    qla2x00_ip_send_logout_port_iocb(ha, pIpDevice);
                    qla2x00_free_loopid(ha, pIpDevice->wLoopId);

                    /* Move device to top of list */
                    qla2x00_free_ip_block(ha, pIpDevice);
                    status = qla2x00_reserve_ip_block(ha, &pIpDevice);
                    break;
                }
            }
        }
        if (status != QL_STATUS_SUCCESS)
            return(status);
    }

    /* Save IP world wide port name */
    *(uint32_t *)(&(pIpDevice->acWorldWideName[0])) = *(uint32_t *)(&(pPortName[0]));
    *(uint32_t *)(&(pIpDevice->acWorldWideName[4])) = *(uint32_t *)(&(pPortName[4]));

    if (wLoopId != PUBLIC_LOOP_DEVICE)
    {
        /* Private loop device */
        pIpDevice->wLoopId = wLoopId;
        pIpDevice->lDeviceFlags = IP_DEV_FLAG_PRESENT;

        DEBUG(printk("qla2x00_add_new_ip_device: "
        "WWN:%02x%02x%02x%02x%02x%02x%02x%02x, LoopID:%x\n\r",
        pIpDevice->acWorldWideName[0], pIpDevice->acWorldWideName[1],
        pIpDevice->acWorldWideName[2], pIpDevice->acWorldWideName[3],
        pIpDevice->acWorldWideName[4], pIpDevice->acWorldWideName[5],
        pIpDevice->acWorldWideName[6], pIpDevice->acWorldWideName[7],
        pIpDevice->wLoopId);)
    }
    else
    {
        /* Public device */
        /* Reserve public loop ID, save it in database */
        if ((status = qla2x00_reserve_loopid(ha, &pIpDevice->wLoopId)) ==
             QL_STATUS_RESOURCE_ERROR)
        {
            if (bForceAdd)
            { 
                IP_DEVICE_BLOCK     *pIpDeviceBump;

                /* Out of loop IDs, bump public device at bottom of list */
                DEBUG(printk("qla2x00_add_new_ip_device: bump device from IP list\n\r");)

                for (pIpDeviceBump = ha->pIpDeviceBottom; pIpDeviceBump != NULL;
                     pIpDeviceBump = pIpDeviceBump->pLastIpDevice)
                {
                    if (pIpDeviceBump->lDeviceFlags & IP_DEV_FLAG_PUBLIC_DEVICE)
                    {
                        /* Do fabric logout, steal loop ID, free bumped IP block */
                        qla2x00_ip_send_logout_port_iocb(ha, pIpDeviceBump);
                        pIpDevice->wLoopId = pIpDeviceBump->wLoopId;
                        qla2x00_free_ip_block(ha, pIpDeviceBump);
                        status = QL_STATUS_SUCCESS;
                        break;
                    }
                }
            }

            if (status != QL_STATUS_SUCCESS)
            {
                /* Failed to get loop ID */
                DEBUG(printk("qla2x00_add_new_ip_device: failed to get loop ID\n\r");)
                qla2x00_free_ip_block(ha, pIpDevice);
                return(status);
            }
        }

        /* Save device data */
        pIpDevice->acPortId[0] = pPortId[0];
        pIpDevice->acPortId[1] = pPortId[1];
        pIpDevice->acPortId[2] = pPortId[2];
        pIpDevice->lDeviceFlags = IP_DEV_FLAG_PUBLIC_DEVICE;

        /* Login public device */
        status = qla2x00_ip_send_login_port_iocb(ha, pIpDevice);

        if (status == QL_STATUS_SUCCESS)
        {
            DEBUG(printk("qla2x00_add_new_ip_device: "
            "WWN:%02x%02x%02x%02x%02x%02x%02x%02x, LoopID:%x, PortID:%x\n\r",
            pIpDevice->acWorldWideName[0], pIpDevice->acWorldWideName[1],
            pIpDevice->acWorldWideName[2], pIpDevice->acWorldWideName[3],
            pIpDevice->acWorldWideName[4], pIpDevice->acWorldWideName[5],
            pIpDevice->acWorldWideName[6], pIpDevice->acWorldWideName[7],
            pIpDevice->wLoopId,
            pIpDevice->acPortId[2]<<16 | pIpDevice->acPortId[1]<<8 | pIpDevice->acPortId[0]);)
        }
        else
        {
            /* Login failed, return resources */
            qla2x00_free_loopid(ha, pIpDevice->wLoopId);
            qla2x00_free_ip_block(ha, pIpDevice);
        }
    }

    return(status);
} /* qla2x00_add_new_ip_device */


/************************************************************************/
/* qla2x00_convert_to_arp                                               */
/*    This function is called to convert an IP send packet to an ARP.   */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pSendCB - pointer to send control block                           */
/*                                                                      */
/* Return Value:                                                        */
/*    TRUE - If converted OK                                            */
/*    FALSE - If unknown packet type                                    */
/************************************************************************/

static int qla2x00_convert_to_arp(scsi_qla_host_t *ha, SEND_CB *pSendCB)
{
    struct sk_buff  *skb = pSendCB->skb;
    PACKET_HEADER   *pPacket = &pSendCB->header;
    ARP_HEADER      *pArpHeader = (ARP_HEADER *)skb->data;
    IP_HEADER       *pIpHeader  = (IP_HEADER *)skb->data;
    int     i;
    uint8_t *pNodeName;

    DEBUG(printk("qla2x00_convert_to_arp: convert packet to ARP\n\r");)

    if (pPacket->sSnapHeader.wSnapPID == SNAP_PID_IP_TYPE)
    {
        /* Convert IP packet to ARP packet */
        pPacket->sNetworkHeader.wDestinationNAA      = NAA_IEEE_MAC_TYPE;
        pPacket->sNetworkHeader.wDestinationAddrHigh = BROADCAST_ADDRESS_WORD;
        pPacket->sNetworkHeader.lDestinationAddrLow  = BROADCAST_ADDRESS;
        pPacket->sSnapHeader.wSnapPID                = SNAP_PID_ARP_TYPE;

        /* Move IP addresses first in reverse, before they get wiped out */
        pArpHeader->acTargetIp[3] = pIpHeader->acDestinationIpAddress[3];
        pArpHeader->acTargetIp[2] = pIpHeader->acDestinationIpAddress[2];
        pArpHeader->acTargetIp[1] = pIpHeader->acDestinationIpAddress[1];
        pArpHeader->acTargetIp[0] = pIpHeader->acDestinationIpAddress[0];
        pArpHeader->acSenderIp[3] = pIpHeader->acSourceIpAddress[3];
        pArpHeader->acSenderIp[2] = pIpHeader->acSourceIpAddress[2];
        pArpHeader->acSenderIp[1] = pIpHeader->acSourceIpAddress[1];
        pArpHeader->acSenderIp[0] = pIpHeader->acSourceIpAddress[0];

        pArpHeader->wHardwareType = ARP_HARDWARE_TYPE;
        pArpHeader->wProtocolType = ARP_PROTOCAL_TYPE;
        pArpHeader->bHLen         = ARP_HLEN;
        pArpHeader->bPLen         = ARP_PLEN;
        pArpHeader->wOperation    = ARP_OPERATION;

        pNodeName = (uint8_t *)&pPacket->sNetworkHeader.wSourceAddrHigh;
        for (i = 0; i < 6; i++, pNodeName++)
        {
            pArpHeader->acSenderHA[i] = *pNodeName;
            pArpHeader->acTargetHA[i] = 0;
        }

        skb->len = sizeof(ARP_HEADER);

        return(TRUE);
    }
    else
    {
        return(FALSE);
    }
} /* qla2x00_convert_to_arp */


/************************************************************************/
/* qla2x00_free_ip_block                                                */
/*    This routine removes an IP device block from active IP list and   */
/*    returns it to the free list.                                      */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pIpDevice - pointer to IP device block to free                    */
/*                                                                      */
/* Return Value:                                                        */
/*    none                                                              */
/************************************************************************/

static void qla2x00_free_ip_block(scsi_qla_host_t *ha,
                                  IP_DEVICE_BLOCK *pIpDevice)
{
    /* Unhook IP device block from active list */
    if (pIpDevice->pLastIpDevice == NULL)
        ha->pIpDeviceTop = pIpDevice->pNextIpDevice;
    else
        ((PIP_DEVICE_BLOCK)pIpDevice->pLastIpDevice)->pNextIpDevice =
            pIpDevice->pNextIpDevice;

    if (pIpDevice->pNextIpDevice == NULL)
        ha->pIpDeviceBottom = pIpDevice->pLastIpDevice;
    else
        ((PIP_DEVICE_BLOCK)pIpDevice->pNextIpDevice)->pLastIpDevice =
            pIpDevice->pLastIpDevice;

    /* Add IP device block to free list */
    pIpDevice->pNextIpDevice = ha->pIpDeviceFree;
    ha->pIpDeviceFree = pIpDevice;
} /* qla2x00_free_ip_block */


/************************************************************************/
/* qla2x00_get_ip_loopid                                                */
/*    This function is called by QLSendLanCommandToIsp to get the loop  */
/*    ID for the specified world wide name. This routine will also      */
/*    update the 1st 2 bytes of the passed in WWN to match the port name*/
/*    stored in the IP database.                                        */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pNodeName - Pointer to World Wide Name                            */
/*    pLoopId - Pointer for returning loop ID                           */
/*                                                                      */
/* Return Value:                                                        */
/*    TRUE - If loop ID returned                                        */
/*    FALSE - If device not logged in and unable to login device        */
/************************************************************************/

static int qla2x00_get_ip_loopid(scsi_qla_host_t *ha,
                                 uint8_t *pNodeName,
                                 uint8_t *pLoopId)
{
    IP_DEVICE_BLOCK     *pIpDevice;

    /* Scan list of logged in IP devices for match */
    for (pIpDevice = ha->pIpDeviceTop; pIpDevice != NULL;
         pIpDevice = pIpDevice->pNextIpDevice)
    {
        if (*(uint16_t *)(&pIpDevice->acWorldWideName[2]) ==
            *(uint16_t *)(pNodeName + 2) &&
            *(uint32_t *)(&pIpDevice->acWorldWideName[4]) ==
            *(uint32_t *)(pNodeName + 4))
        {
            /* Found match, return loop ID  */
            *pLoopId = (uint8_t)pIpDevice->wLoopId;

            /* Set first 2 bytes of port name */
            *((uint16_t *)pNodeName) = *((uint16_t *)(&pIpDevice->acWorldWideName[0]));

            if (pIpDevice != ha->pIpDeviceTop)
            {
                /* Device not at top, move it to top of list */
                /* Unhook it first */
                if (pIpDevice == ha->pIpDeviceBottom)
                {
                    ha->pIpDeviceBottom = pIpDevice->pLastIpDevice;
                    ((PIP_DEVICE_BLOCK)pIpDevice->pLastIpDevice)->pNextIpDevice = NULL;
                }
                else
                {
                    ((PIP_DEVICE_BLOCK)pIpDevice->pLastIpDevice)->pNextIpDevice = pIpDevice->pNextIpDevice;
                    ((PIP_DEVICE_BLOCK)pIpDevice->pNextIpDevice)->pLastIpDevice = pIpDevice->pLastIpDevice;
                }

                /* Now put it at top of list */
                pIpDevice->pNextIpDevice = ha->pIpDeviceTop;
                pIpDevice->pLastIpDevice = NULL;
                ha->pIpDeviceTop->pLastIpDevice = pIpDevice;
                ha->pIpDeviceTop = pIpDevice;
            }
            return(TRUE);
        }
    }

    /* Check for broadcast packet */
    if (*(uint16_t *)(&(pNodeName[2])) == (uint16_t)0xFFFF &&
        *(uint32_t *)(&(pNodeName[4])) == (uint32_t)0xFFFFFFFF)
    {
        /* Broadcast packet, return broadcast loop ID  */
        *pLoopId = BROADCAST;

        /* Set first 2 bytes of port name */
        *((uint16_t *)pNodeName) = NAA_IEEE_MAC_TYPE;
        return(TRUE);
    }

    /* Check for multicast packet */
    if (*(pNodeName+2) & 0x01)
    {
        /* Use broadcast loop ID for multicast packets  */
        *pLoopId = BROADCAST;

        /* Set first 2 bytes of port name */
        *((uint16_t *)pNodeName) = NAA_IEEE_MAC_TYPE;
        return(TRUE);
    }

    /* Try sending FARP IOCB to request login */
    /* TBD */

    DEBUG(printk("qla2x00_get_ip_loopid: ID not found for XX XX %02x %02x %02x %02x %02x %02x\n\r",
    *(pNodeName+2),*(pNodeName+3),*(pNodeName+4),*(pNodeName+5),*(pNodeName+6),*(pNodeName+7));)
    return(FALSE);
} /* qla2x00_get_ip_loopid */


/************************************************************************/
/* qla2x00_reserve_ip_block                                             */
/*    This routine moves an unused IP device block from the free list   */
/*    to the top of the active IP device list.                          */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pIpDevBlk - address of pointer to IP device block                 */
/*                                                                      */
/* Return Value:                                                        */
/*    QL_STATUS_SUCCESS         if no error                             */
/*    QL_STATUS_RESOURCE_ERROR  if out of IP device blocks              */
/************************************************************************/

static int qla2x00_reserve_ip_block(scsi_qla_host_t *ha,
                                    PIP_DEVICE_BLOCK *pIpDevBlk)
{
    IP_DEVICE_BLOCK     *pIpDevice;

    /* Get free IP device block */
    if ((pIpDevice = ha->pIpDeviceFree) != NULL)
    { 
        /* Remove IP device block from free list */
        ha->pIpDeviceFree = pIpDevice->pNextIpDevice;

        /* Add IP device block to top of IP device list */
        pIpDevice->pNextIpDevice = ha->pIpDeviceTop;
        pIpDevice->pLastIpDevice = NULL;
        if (ha->pIpDeviceTop == NULL)
            ha->pIpDeviceBottom = pIpDevice;
        else
            ha->pIpDeviceTop->pLastIpDevice = pIpDevice;
        ha->pIpDeviceTop = pIpDevice;

        *pIpDevBlk = pIpDevice;
        return(QL_STATUS_SUCCESS);
    }
    else
    {
        /* Out of IP blocks */
        DEBUG(printk("qla2x00_reserve_ip_block: out of IP blocks\n\r");)
        return(QL_STATUS_RESOURCE_ERROR);
    }
} /* qla2x00_reserve_ip_block */


/************************************************************************/
/* qla2x00_update_ip_device_data                                        */
/*    This routine searchs for the device node name in the current IP   */
/*    database and updates database.                                    */
/*                                                                      */
/*    If device found:                                                  */
/*     * Handle device movement between public and private loops        */
/*     * Handle device movement between FL ports on switch              */
/*     * Mark device present                                            */
/*     * Log in device if necessary                                     */
/*    If device not found and private loop device:                      */
/*     * Insert the new entry in database                               */
/*    If device not found and public IP device:                         */
/*     * Ignore device until packet received from device                */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pDeviceData                                                       */
/*    NOTE: If pDeviceData->loop_id =  0 - 7E, private loop device      */
/*          If pDeviceData->loop_id = -1, public loop device            */
/*                                                                      */
/* Return Value:                                                        */
/*    QL_STATUS_SUCCESS         if no error                             */
/*    QL_STATUS_RESOURCE_ERROR  if out of loop IDs or IP device blocks  */
/*    QL_STATUS_FATAL_ERROR     if fatal error                          */
/*    QL_STATUS_ERROR           if any other error                      */
/************************************************************************/

static int qla2x00_update_ip_device_data(scsi_qla_host_t *ha,
                                         fcdev_t *pDeviceData)
{
    IP_DEVICE_BLOCK     *pIpDevice;
    int                 status = 0;

    if (!ha->flags.enable_ip)
    {
        /* IP not enabled, just return */
        return(QL_STATUS_SUCCESS);
    }

    /* Scan list of IP devices for match */
    for (pIpDevice = ha->pIpDeviceTop; pIpDevice != NULL;
         pIpDevice = pIpDevice->pNextIpDevice)
    {
        if (*(uint32_t *)(&(pDeviceData->wwn[0])) == *(uint32_t *)(&(pIpDevice->acWorldWideName[0])) &&
            *(uint32_t *)(&(pDeviceData->wwn[4])) == *(uint32_t *)(&(pIpDevice->acWorldWideName[4])))
        {
            /* Found device in IP device list */
            DEBUG(printk("qla2x00_update_ip_device_data: already in IP list, port ID: %x\n\r",
            pIpDevice->acPortId[2]<<16 | pIpDevice->acPortId[1]<<8 | pIpDevice->acPortId[0]);)

            if (pDeviceData->flag != DEV_PUBLIC &&
                !(pIpDevice->lDeviceFlags & IP_DEV_FLAG_PUBLIC_DEVICE))
            {
                /* Device on private loop now, was on private loop before */
                DEBUG(printk("qla2x00_update_ip_device_data: was private loop, now private loop\n\r");)
                /* Update private loop ID in database */
                pIpDevice->wLoopId = pDeviceData->loop_id;
                pIpDevice->lDeviceFlags |= IP_DEV_FLAG_PRESENT;
            }
            else if (pDeviceData->flag != DEV_PUBLIC &&
                     (pIpDevice->lDeviceFlags & IP_DEV_FLAG_PUBLIC_DEVICE))
            {
                /* Device on private loop now, was public device before */
                DEBUG(printk("qla2x00_update_ip_device_data: was public, now private loop\n\r");)
                /* If loop ID changed, logout device and free loop ID */
                if (pDeviceData->loop_id != pIpDevice->wLoopId)
                { 
                    qla2x00_ip_send_logout_port_iocb(ha, pIpDevice);
                    qla2x00_free_loopid(ha, pIpDevice->wLoopId);

                    /* Clear public device flag and save private loop ID in database */
                    pIpDevice->lDeviceFlags &= ~IP_DEV_FLAG_PUBLIC_DEVICE;
                    pIpDevice->wLoopId = pDeviceData->loop_id;
                }
                pIpDevice->lDeviceFlags |= IP_DEV_FLAG_PRESENT;
            }
            else if (pDeviceData->flag == DEV_PUBLIC &&
                     !(pIpDevice->lDeviceFlags & IP_DEV_FLAG_PUBLIC_DEVICE))
            {
                /* Device public now, was on private loop before */
                DEBUG(printk("qla2x00_update_ip_device_data: was private loop, now public\n\r");)

                /* Reserve public loop ID, save it in database */
                if ((status = qla2x00_reserve_loopid(ha, &pIpDevice->wLoopId)) ==
                    QL_STATUS_SUCCESS)
                { 
                    /* Save port ID and set public device flag */
                    pIpDevice->acPortId[0] = pDeviceData->d_id.r.d_id[0];
                    pIpDevice->acPortId[1] = pDeviceData->d_id.r.d_id[1];
                    pIpDevice->acPortId[2] = pDeviceData->d_id.r.d_id[2];
                    pIpDevice->lDeviceFlags |= IP_DEV_FLAG_PUBLIC_DEVICE;

                    /* Login public device */
                    status = qla2x00_ip_send_login_port_iocb(ha, pIpDevice);
                }
                if (status == QL_STATUS_RESOURCE_ERROR)
                {
                    /* Out of loop IDs */
                    pIpDevice->lDeviceFlags &= ~IP_DEV_FLAG_PUBLIC_DEVICE;
                }
            }
            else
            {
                /* Device public now, was public device before */
                DEBUG(printk("qla2x00_update_ip_device_data: was public, now public\n\r");)
                /* Check if port ID changed */
                if (pIpDevice->acPortId[0] != pDeviceData->d_id.r.d_id[0] ||
                    pIpDevice->acPortId[1] != pDeviceData->d_id.r.d_id[1] ||
                    pIpDevice->acPortId[2] != pDeviceData->d_id.r.d_id[2])
                {
                    /* Save new port ID */
                    pIpDevice->acPortId[0] = pDeviceData->d_id.r.d_id[0];
                    pIpDevice->acPortId[1] = pDeviceData->d_id.r.d_id[1];
                    pIpDevice->acPortId[2] = pDeviceData->d_id.r.d_id[2];
                    DEBUG(printk("qla2x00_update_ip_device_data: Port ID changed\n\r");)

                    /* Logout public device */
                    qla2x00_ip_send_logout_port_iocb(ha, pIpDevice);
                }

                /* Login public device */
                status = qla2x00_ip_send_login_port_iocb(ha, pIpDevice);

                if (status == QL_STATUS_RESOURCE_ERROR)
                {
                    /* Out of loop IDs */
                    pIpDevice->lDeviceFlags &= ~IP_DEV_FLAG_PUBLIC_DEVICE;
                }
            }
            return(status);
        }
    }

    /* Device not found in database */
    DEBUG(printk("qla2x00_update_ip_device_data: device NOT in IP list\n\r");)

    /* If private loop device, add device to IP list */
    /* Public devices will be added as needed when packet received */
    if (pDeviceData->flag != DEV_PUBLIC)
    {
        /* Add (force) new private loop device to IP list */
        status = qla2x00_add_new_ip_device(ha, pDeviceData->loop_id,
                                   NULL,
                                   (uint8_t *)&pDeviceData->wwn[0], TRUE);
    }

    /* The following code is temporary, until FARP supported */
    /* Login all IP public devices for now */
    if (pDeviceData->flag == DEV_PUBLIC)
    {
        /* Add (don't force) new public device to IP list */
        status = qla2x00_add_new_ip_device(ha, PUBLIC_LOOP_DEVICE,
                                   (uint8_t *)&pDeviceData->d_id,
                                   (uint8_t *)&pDeviceData->wwn[0], FALSE);
    }
    return(status);
} /* qla2x00_update_ip_device_data */


/************************************************************************/
/* qla2x00_reserve_loopid                                               */
/*    This routine reserves an unused public loop ID.                   */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pLoopID   - pointer to word for returning loop ID                 */
/*                                                                      */
/* Return Value:                                                        */
/*    QL_STATUS_SUCCESS         if no error                             */
/*    QL_STATUS_RESOURCE_ERROR  if out of loop IDs                      */
/************************************************************************/

static int qla2x00_reserve_loopid(scsi_qla_host_t *ha, uint16_t *pLoopId)
{
    int     i;

    /* Look for unused loop ID */
    for (i = ha->min_external_loopid; i < ha->max_public_loop_ids; i++)
    {
        if (!ha->fabricid[i].in_use)
        {
            /* Found free loop ID */
            ha->fabricid[i].in_use = TRUE;
            *pLoopId = i;

            DEBUG(printk("qla2x00_reserve_loopid: assigned loop ID %x\n\r", *pLoopId);)
            return QL_STATUS_SUCCESS;
        }
    }

    /* Out of loop IDs */
    *pLoopId = ha->max_public_loop_ids + 1;     /* Set out of range */
    DEBUG(printk("qla2x00_reserve_loopid: out of loop IDs\n\r");)
    return QL_STATUS_RESOURCE_ERROR;
} /* qla2x00_reserve_loopid */


/************************************************************************/
/* qla2x00_free_loopid                                                  */
/*    This routine frees a public loop ID.                              */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    wLoopID   - loop ID to free                                       */
/*                                                                      */
/* Return Value:                                                        */
/*    none                                                              */
/************************************************************************/

static void qla2x00_free_loopid(scsi_qla_host_t *ha, uint16_t wLoopId)
{
    if (wLoopId < ha->max_public_loop_ids)
    {
        ha->fabricid[wLoopId].in_use = FALSE;
        DEBUG(printk("qla2x00_free_loopid: free loop ID %x\n\r", wLoopId);)
    }
    else
    {
        DEBUG(printk("qla2x00_free_loopid: loop ID %x out of range\n\r", wLoopId);)
    }
} /* qla2x00_free_loopid */


/************************************************************************/
/* qla2x00_ip_send_login_port_iocb                                      */
/*    This routine is call to build and send a mailbox IOCB for logging */
/*    in a fabric port.                                                 */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pIpDevice - pointer to IP device block.                           */
/*                                                                      */
/* Return Value:                                                        */
/*    QL_STATUS_SUCCESS         if no error                             */
/*    QL_STATUS_ERROR           if any other error                      */
/************************************************************************/

static int qla2x00_ip_send_login_port_iocb(scsi_qla_host_t *ha,
                                           IP_DEVICE_BLOCK *pIpDevice)
{
	unsigned long flags = 0;
    PMBX_IOCB   pIocb;

    DEBUG(printk("qla2x00_ip_send_login_port_iocb: port ID: %x\n",
          pIpDevice->acPortId[2]<<16 | pIpDevice->acPortId[1]<<8 | pIpDevice->acPortId[0]);)

    /* Send marker if required */
    if (ha->marker_needed != 0) {
        if(qla2x00_marker(ha, 0, 0, MK_SYNC_ALL) != QLA2X00_SUCCESS) {
            return QL_STATUS_ERROR;
        }
        ha->marker_needed = 0;
    }

	spin_lock_irqsave(&ha->hardware_lock, flags);

	pIocb = (PMBX_IOCB)qla2x00_req_pkt(ha);
	if (pIocb == NULL) {
        DEBUG(printk("qla2x00_ip_send_login_port_iocb: failed\n");)
		spin_unlock_irqrestore(&ha->hardware_lock, flags);
		return QL_STATUS_ERROR;
	}

    /* Build fabric login MBX IOCB */
    pIocb->Header   = ET_MAILBOX_COMMAND | (1 << 8) | (SOURCE_IP << 16);
    pIocb->Handle   = pIpDevice->wDeviceIndex | (MBC_LOGIN_FABRIC_PORT << 16);
    pIocb->LoopId   = pIpDevice->wLoopId;
    pIocb->Mailbox0 = MBC_LOGIN_FABRIC_PORT;
    pIocb->Mailbox1 = (pIpDevice->wLoopId << 8) | 
                      (MBC_NO_PROCESS_LOGIN | MBC_NO_PLOGI_IF_LOGGED_IN);
    pIocb->Mailbox2 = pIpDevice->acPortId[2];
    pIocb->Mailbox3 = (pIpDevice->acPortId[1] << 8) | pIpDevice->acPortId[0];
    pIocb->Mailbox6 = 0;
    pIocb->Mailbox7 = 0;

	/* Issue command to ISP */
	qla2x00_isp_cmd(ha);

	spin_unlock_irqrestore(&ha->hardware_lock, flags);

    return QL_STATUS_SUCCESS;
} /* qla2x00_ip_send_login_port_iocb */


/************************************************************************/
/* qla2x00_ip_send_logout_port_iocb                                     */
/*    This routine is call to build and send a mailbox IOCB for logging */
/*    out a fabric port.                                                */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pIpDevice - pointer to IP device block.                           */
/*                                                                      */
/* Return Value:                                                        */
/*    QL_STATUS_SUCCESS         if no error                             */
/*    QL_STATUS_ERROR           if any other error                      */
/************************************************************************/

static int qla2x00_ip_send_logout_port_iocb(scsi_qla_host_t *ha,
                                           IP_DEVICE_BLOCK *pIpDevice)
{
	unsigned long flags = 0;
    PMBX_IOCB   pIocb;

    DEBUG(printk("qla2x00_ip_send_logout_port_iocb: port ID: %x\n",
          pIpDevice->acPortId[2]<<16 | pIpDevice->acPortId[1]<<8 | pIpDevice->acPortId[0]);)

    /* Send marker if required */
    if (ha->marker_needed != 0) {
        if(qla2x00_marker(ha, 0, 0, MK_SYNC_ALL) != QLA2X00_SUCCESS) {
            return QL_STATUS_ERROR;
        }
        ha->marker_needed = 0;
    }

	spin_lock_irqsave(&ha->hardware_lock, flags);

	pIocb = (PMBX_IOCB)qla2x00_req_pkt(ha);
	if (pIocb == NULL) {
        DEBUG(printk("qla2x00_ip_send_logout_port_iocb: failed\n");)
		spin_unlock_irqrestore(&ha->hardware_lock, flags);
		return QL_STATUS_ERROR;
	}

    /* Build fabric logout MBX IOCB */
    pIocb->Header   = ET_MAILBOX_COMMAND | (1 << 8) | (SOURCE_IP << 16);
    pIocb->Handle   = pIpDevice->wDeviceIndex | (MBC_LOGOUT_FABRIC_PORT << 16);
    pIocb->LoopId   = pIpDevice->wLoopId;
    pIocb->Mailbox0 = MBC_LOGOUT_FABRIC_PORT;
    pIocb->Mailbox1 = pIpDevice->wLoopId << 8;
    pIocb->Mailbox2 = 0;
    pIocb->Mailbox3 = 0;
    pIocb->Mailbox6 = 0;
    pIocb->Mailbox7 = 0;

	/* Issue command to ISP */
	qla2x00_isp_cmd(ha);

	spin_unlock_irqrestore(&ha->hardware_lock, flags);

    return QL_STATUS_SUCCESS;
} /* qla2x00_ip_send_logout_port_iocb */


/************************************************************************/
/* qla2x00_ip_mailbox_iocb_done                                         */
/*    This routine is call to process a MBX IOCB completion.            */
/*    Note: this is only used for fabric login and logout for now.      */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pIocb - pointer completed MBX IOCB.                               */
/*                                                                      */
/* Return Value:                                                        */
/*    none                                                              */
/************************************************************************/

static void qla2x00_ip_mailbox_iocb_done(scsi_qla_host_t *ha,
                                         PMBX_IOCB  pIocb)
{
    PIP_DEVICE_BLOCK    pIpDevice;
    int     index;

    DEBUG(printk("qla2x00_ip_mailbox_iocb_done: cmd %x, status %x, MB0 %x, MB1 %x, MB2 %x\n",
          pIocb->Handle >> 16, pIocb->Status, pIocb->Mailbox0, pIocb->Mailbox1, pIocb->Mailbox2);)

    /* Get device block pointer */
    index = pIocb->Handle & 0xffff;
    if (index >= QLLAN_MAX_IP_DEVICES)
    {
        /* Bad handle from ISP */
        DEBUG(printk("qla2x00_ip_mailbox_iocb_done: bad handle from isp\n");)
        /* TBD */
        return;
    }
    pIpDevice = &ha->asIpDeviceDatabase[index];

    if ((pIocb->Handle >> 16) == MBC_LOGOUT_FABRIC_PORT)
    {
        /* Check fabric logout completion status */
        if ( /*pIocb->Status == CS_COMPLETE && */
            pIocb->Mailbox0 == MBS_COMMAND_COMPLETE)
        {
            /* Logout successful */
        }
        else
        {
            DEBUG(printk("qla2x00_ip_mailbox_iocb_done: fabric logout failed\n");)
        }
    }
    else
    {
        /* Check fabric login completion status */
        /* Note: sometimes ISP returns Status=0x30 and MB0=0x4000 */
        /* Therefore, only check MB0 for now */
        if ( /* pIocb->Status == CS_COMPLETE && */
            pIocb->Mailbox0 == MBS_COMMAND_COMPLETE)
        {
            /* Login successful */
            pIpDevice->lDeviceFlags |= IP_DEV_FLAG_PRESENT;
        }
        else if (pIocb->Mailbox0 == MBS_PORT_ID_IN_USE)
        {
            /* Different loop ID already assigned to port ID */
            /* Use the one that is already assigned */
            qla2x00_free_loopid(ha, pIpDevice->wLoopId);
            pIpDevice->wLoopId = pIocb->Mailbox1;

            /* Do logout first and then relogin */
            qla2x00_ip_send_logout_port_iocb(ha, pIpDevice);
            qla2x00_ip_send_login_port_iocb(ha, pIpDevice);
        }
        else if (pIocb->Mailbox0 == MBS_LOOP_ID_IN_USE)
        {
            /* Loop ID already used for different port ID */
            /* Get a new loop ID and reissue login request */
            if (qla2x00_reserve_loopid(ha, &pIpDevice->wLoopId) == QL_STATUS_SUCCESS)
            {
                qla2x00_ip_send_login_port_iocb(ha, pIpDevice);
            }
            else
            {
                DEBUG(printk("qla2x00_ip_mailbox_iocb_done: out of loop IDs\n");)
                qla2x00_free_ip_block(ha, pIpDevice);
            }
        }
        else
        {
            /* Login failed, return resources */
            DEBUG(printk("qla2x00_ip_mailbox_iocb_done: fabric login failed\n");)
            qla2x00_free_loopid(ha, pIpDevice->wLoopId);
            qla2x00_free_ip_block(ha, pIpDevice);
        }
    }
} /* qla2x00_ip_mailbox_iocb_done */


/************************************************************************/
/* qla2x00_send_packet                                                  */
/*    This routine is called from the IP network driver to pass IP      */
/*    packet to the ISP for transmission.                               */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pSendCB - pointer to send control block.                          */
/*                                                                      */
/* Return Value:                                                        */
/*    QL_STATUS_SUCCESS         if packet sent                          */
/*    QL_STATUS_RESOURCE_ERROR  if too busy to send now                 */
/*    QL_STATUS_ERROR           if any other error                      */
/************************************************************************/

int qla2x00_send_packet(scsi_qla_host_t *ha, SEND_CB *pSendCB)
{
	unsigned long flags = 0;
    struct sk_buff  *skb = pSendCB->skb;
    device_reg_t    *reg   = ha->iobase;
    uint16_t        cnt;
    uint32_t        lTagVal;
    int             i;
    PIP_COMMAND_ENTRY   pCmdEntry;
  	u_long          phys_address;

    DEBUG3(printk("qla2x00_send_packet: enter\n\r");)

    /* Check adapter state */
    if (!ha->flags.online)
    {
        return QL_STATUS_ERROR;
    }

    /* Send marker if required */
    if (ha->marker_needed != 0) {
        if(qla2x00_marker(ha, 0, 0, MK_SYNC_ALL) != QLA2X00_SUCCESS) {
            return QL_STATUS_ERROR;
        }
        ha->marker_needed = 0;
    }

    /* Acquire ring specific lock */
	spin_lock_irqsave(&ha->hardware_lock, flags);

    if (ha->req_q_cnt < 4)
    {
        /* Update number of free request entries */
#if defined(ISP2100) || defined(ISP2200)
        cnt = qla2x00_debounce_register(&reg->mailbox4);
#else
        cnt = qla2x00_debounce_register(&reg->req_q_out);
#endif

        if (ha->req_ring_index < cnt)
            ha->req_q_cnt = cnt - ha->req_ring_index;
        else
            ha->req_q_cnt = REQUEST_ENTRY_CNT - (ha->req_ring_index - cnt);
    }

    if (ha->req_q_cnt >= 4)
    {
        /* Get tag handle for command */
        lTagVal = ha->lCurrentTagIndex;
        for (i = 0; i < MAX_SEND_PACKETS; i++)
        {
            lTagVal++;
            if (lTagVal == MAX_SEND_PACKETS)
            {
                lTagVal = 0;
            }
            if (ha->apActiveIpQueue[lTagVal] == NULL)
            {
                ha->lCurrentTagIndex = lTagVal;
                goto FoundTagHandle;
            }
        }
    }

    /* Low on resources, try again later */
	spin_unlock_irqrestore(&ha->hardware_lock, flags);
    return QL_STATUS_RESOURCE_ERROR;

FoundTagHandle:

    /* Build ISP command packet */
    pCmdEntry = (PIP_COMMAND_ENTRY)ha->request_ring_ptr;
    pCmdEntry->lHeader      = ET_IP_COMMAND_64 | 1 << 8;
    pCmdEntry->lHandle      = lTagVal;
    pCmdEntry->cReserved1   = 0;

    /* Get destination loop ID for packet */
    if (!qla2x00_get_ip_loopid(ha, (uint8_t *)&pSendCB->header, &pCmdEntry->cLoopId))
    {
        /* Failed to get loop ID, convert packet to ARP */
        if (qla2x00_convert_to_arp(ha, pSendCB))
        {
            /* Broadcast ARP */
            pCmdEntry->cLoopId = BROADCAST;
        }
        else
        {
            /* Return packet */
        	spin_unlock_irqrestore(&ha->hardware_lock, flags);
            return QL_STATUS_ERROR;
        }
    }

    /* Set command timeout value */
    pCmdEntry->wTimeout = 5;

    /* Set data direction control flags */
    pCmdEntry->wControlFlags = CF_WRITE;

    pCmdEntry->wReserved2    = 0;
    pCmdEntry->wServiceClass = 0;

    pCmdEntry->wSegmentCount = 2;

    pCmdEntry->ds.asDataSeg64[0].lBase[0] = pSendCB->physicalAddressLow;
    pCmdEntry->ds.asDataSeg64[0].lBase[1] = pSendCB->physicalAddressHigh;
    pCmdEntry->ds.asDataSeg64[0].lCount = sizeof(PACKET_HEADER);

    phys_address = virt_to_bus(skb->data);
    pCmdEntry->ds.asDataSeg64[1].lBase[0] = LS_64BITS(phys_address);
    pCmdEntry->ds.asDataSeg64[1].lBase[1] = MS_64BITS(phys_address);
    pCmdEntry->ds.asDataSeg64[1].lCount = skb->len;

    pCmdEntry->lTotalByteCount = skb->len + sizeof(PACKET_HEADER);

    /* Adjust ring index. */
    ha->req_ring_index++;
    if (ha->req_ring_index == REQUEST_ENTRY_CNT)
    {
        ha->req_ring_index = 0;
        ha->request_ring_ptr = ha->request_ring;
    } else
        ha->request_ring_ptr++;

    ha->req_q_cnt--;
    ha->apActiveIpQueue[lTagVal] = pSendCB;

    /* Set chip new ring index. */
#if defined(ISP2100) || defined(ISP2200)
    WRT_REG_WORD(&reg->mailbox4, ha->req_ring_index);
#else
    WRT_REG_WORD(&reg->req_q_in, ha->req_ring_index);
#endif

	spin_unlock_irqrestore(&ha->hardware_lock, flags);
    return QL_STATUS_SUCCESS;
} /* qla2x00_send_packet */


/************************************************************************/
/* qla2x00_ip_send_complete                                             */
/*    This routine handles IP send completion.                          */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    lTagVal - index into active IP queue                              */
/*                                                                      */
/* Return Value:                                                        */
/*    none                                                              */
/************************************************************************/

static void qla2x00_ip_send_complete(scsi_qla_host_t *ha, uint32_t lTagVal)
{
	SEND_CB      *pSendCB;

    DEBUG3(printk("qla2x00_ip_send_complete: enter\n\r");)

	/* Set packet pointer from queue entry handle */
	if (lTagVal < MAX_SEND_PACKETS)
	{
		if ((pSendCB = (SEND_CB *)ha->apActiveIpQueue[lTagVal]) != NULL)
		{
			ha->apActiveIpQueue[lTagVal] = NULL;

			/* Return send packet to IP driver */
			(*ha->pSendCompletionRoutine) (pSendCB);
            return;
		}
	}

    /* Invalid handle from RISC, reset RISC firmware */
    printk(KERN_WARNING "qla2x00: bad IP send handle %x\n", lTagVal);
    set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
} /* qla2x00_ip_send_complete */


/************************************************************************/
/* qla2x00_ip_receive                                                   */
/*    This routine handles an incoming IP receive IOCB.                 */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    pkt - pointer to IP receive IOCB                                  */
/*                                                                      */
/* Return Value:                                                        */
/*    none                                                              */
/************************************************************************/

static void qla2x00_ip_receive(scsi_qla_host_t *ha, response_t *pkt)
{
	PIP_RECEIVE_ENTRY pIpReceiveEntry = (PIP_RECEIVE_ENTRY) pkt;
	PBUFFER_CB      pBufferCB, pNextBufferCB;
	uint32_t        lTagVal;
	uint32_t        lPacketSize;
	uint16_t        wBufferCount;
	uint32_t        lReceiveBufferSize;

    DEBUG3(printk("qla2x00_ip_receive: enter\n\r");)

	/* If split buffer, set header size for 1st buffer */
	if (pIpReceiveEntry->wCompletionStatus & IP_REC_STATUS_SPLIT_BUFFER)
		lReceiveBufferSize = ha->wHeaderSize;
	else
		lReceiveBufferSize = ha->lReceiveBufferSize;

	if ((lTagVal = pIpReceiveEntry->waBufferHandle[0]) >= ha->wReceiveBufferCount)
	{
		/* Invalid handle from RISC, reset RISC firmware */
		printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
		set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		return;
	}
	pBufferCB = &ha->pReceiveBufferCBs[lTagVal];

	if (!(pBufferCB->lFlags & BCB_FLAGS_RISC_OWNS_BUFFER))
	{
		/* Invalid handle from RISC, reset RISC firmware */
		printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
		set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		return;
	}

	/* Set buffer belongs to driver now */
	pBufferCB->lFlags &= ~BCB_FLAGS_RISC_OWNS_BUFFER;

	lPacketSize = pIpReceiveEntry->wSequenseLength;
	pBufferCB->lPacketSize = lPacketSize;
	pNextBufferCB = pBufferCB;

	for (wBufferCount = 1;; wBufferCount++)
	{
		if (lPacketSize > lReceiveBufferSize)
		{
			pNextBufferCB->lBufferSize = lReceiveBufferSize;
			lPacketSize -= lReceiveBufferSize;
			/* If split buffer, only use header size on 1st buffer */
			lReceiveBufferSize = ha->lReceiveBufferSize;

			if ((lTagVal = pIpReceiveEntry->waBufferHandle[wBufferCount]) >=
			    ha->wReceiveBufferCount)
			{
				/* Invalid handle from RISC, reset RISC firmware */
				printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
				set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
				return;
			}
			pNextBufferCB->pNextBufferCB = &ha->pReceiveBufferCBs[lTagVal];
			pNextBufferCB = pNextBufferCB->pNextBufferCB;

			if (!(pNextBufferCB->lFlags & BCB_FLAGS_RISC_OWNS_BUFFER))
			{
				/* Invalid handle from RISC, reset RISC firmware */
				printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
				set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
				return;
			}

			/* Set buffer belongs to driver now */
			pNextBufferCB->lFlags &= ~BCB_FLAGS_RISC_OWNS_BUFFER;
		}
		else
		{
			pNextBufferCB->lBufferSize = lPacketSize;
			pNextBufferCB->pNextBufferCB = NULL;
			break;
		}
	}

	/* Check for incoming ARP packet with matching IP address */
	if (pIpReceiveEntry->wServiceClass == 0)
	{
		PPACKET_HEADER pPacket = (PPACKET_HEADER) pBufferCB->pBuffer;
		PIP_DEVICE_BLOCK pIpDevice;
		uint8_t acPortId[3];

		/* Scan list of IP devices to see if login needed */
		for (pIpDevice = ha->pIpDeviceTop; pIpDevice != NULL;
		     pIpDevice = pIpDevice->pNextIpDevice)
		{
			if (*(uint16_t *) (&pIpDevice->acWorldWideName[2]) ==
			    pPacket->sNetworkHeader.wSourceAddrHigh &&
			    *(uint32_t *) (&pIpDevice->acWorldWideName[4]) ==
			    pPacket->sNetworkHeader.lSourceAddrLow)
			{
				/* Device already in IP list, skip login */
				goto SkipDeviceLogin;
			}
		}

		/* Device not in list, need to do login */
		acPortId[2] = pIpReceiveEntry->cS_IDHigh;
		acPortId[1] = (uint8_t) (pIpReceiveEntry->wS_IDLow >> 8);
		acPortId[0] = (uint8_t) pIpReceiveEntry->wS_IDLow;

		/* Make sure its not a local device */
		if (acPortId[2] == ha->d_id.b.domain && acPortId[1] == ha->d_id.b.area)
		{
			goto SkipDeviceLogin;
		}

		if (qla2x00_add_new_ip_device(ha, PUBLIC_LOOP_DEVICE,
					      acPortId,
					      (uint8_t *) & pPacket->sNetworkHeader.wSourceNAA,
					      TRUE) == QL_STATUS_FATAL_ERROR)
        {
			/* Fatal error, reinitialize */
			set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
		}
	}

SkipDeviceLogin:

	/* Pass received packet to IP driver */
	pBufferCB->wBufferCount = wBufferCount;
	(*ha->pReturnReceivePacketsRoutine)
	    (ha->pReturnReceivePacketsContext, pBufferCB);

	/* Keep track of RISC buffer pointer (for IP reinit) */
	ha->wIpBufferOut += wBufferCount;
	if (ha->wIpBufferOut >= IP_BUFFER_QUEUE_DEPTH)
		ha->wIpBufferOut -= IP_BUFFER_QUEUE_DEPTH;
} /* qla2x00_ip_receive */


/************************************************************************/
/* qla2x00_ip_receive_fastpost                                          */
/*    This routine handles an incoming IP receive fast post.            */
/*                                                                      */
/* Arguments:                                                           */
/*    ha - adapter block pointer                                        */
/*    type - type of fast post                                          */
/*                                                                      */
/* Return Value:                                                        */
/*    none                                                              */
/************************************************************************/

static void qla2x00_ip_receive_fastpost(scsi_qla_host_t *ha, uint16_t type)
{
    device_reg_t    *reg   = ha->iobase;
    volatile uint16_t *pNextMailbox;
	PBUFFER_CB      pBufferCB, pNextBufferCB;
	uint32_t        lTagVal;
	uint32_t        lPacketSize;
	uint16_t        wBufferCount;
	uint32_t        lReceiveBufferSize;

    DEBUG3(printk("qla2x00_ip_receive_fastpost: enter\n\r");)

    pNextMailbox = &reg->mailbox10;

    /* If split buffer, set header size for 1st buffer */
    if (type == MBA_IP_RECEIVE_COMPLETE_SPLIT)
        lReceiveBufferSize = ha->wHeaderSize;
    else
        lReceiveBufferSize = ha->lReceiveBufferSize;

    if ((lTagVal = RD_REG_WORD(pNextMailbox)) >= ha->wReceiveBufferCount)
    {
        goto InvalidIpBufferHandle;
    }

    pBufferCB = &ha->pReceiveBufferCBs[lTagVal];

    if (!(pBufferCB->lFlags &  BCB_FLAGS_RISC_OWNS_BUFFER))
    {
        goto InvalidIpBufferHandle;
    }

    /* Set buffer belongs to driver now */
    pBufferCB->lFlags &= ~BCB_FLAGS_RISC_OWNS_BUFFER; 

    lPacketSize = RD_REG_WORD(&reg->mailbox3);
    pBufferCB->lPacketSize = lPacketSize;
    pNextBufferCB = pBufferCB;

    for (wBufferCount = 1; ; wBufferCount++)
    {
        if (lPacketSize > lReceiveBufferSize)
        {
            pNextBufferCB->lBufferSize = lReceiveBufferSize;
            lPacketSize               -= lReceiveBufferSize;

            /* If split buffer, only use header size on 1st buffer */
            lReceiveBufferSize = ha->lReceiveBufferSize;

            pNextMailbox++;
            if ((lTagVal = RD_REG_WORD(pNextMailbox)) >=
                ha->wReceiveBufferCount)
            {
InvalidIpBufferHandle:
                printk(KERN_WARNING "qla2x00: bad IP receive fast post handle %x\n", 
                    lTagVal);
                set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
                return;
            }

            pNextBufferCB->pNextBufferCB = &ha->pReceiveBufferCBs[lTagVal];
            pNextBufferCB = pNextBufferCB->pNextBufferCB;

            if (!(pNextBufferCB->lFlags & BCB_FLAGS_RISC_OWNS_BUFFER))
            {
                goto InvalidIpBufferHandle;
            }

            /* Set buffer belongs to driver now */
            pNextBufferCB->lFlags &= ~BCB_FLAGS_RISC_OWNS_BUFFER; 
        }
        else
        {
            pNextBufferCB->lBufferSize = lPacketSize;
            pNextBufferCB->pNextBufferCB = NULL;
            break;
        }
    }

    /* Clear interrupt */
    WRT_REG_WORD(&reg->host_cmd, HC_CLR_RISC_INT);
#if defined(ISP2100) || defined(ISP2200)
    WRT_REG_WORD(&reg->semaphore, 0);
#endif

    /* Pass received packet to IP driver */
    pBufferCB->wBufferCount = wBufferCount;

    (*ha->pReturnReceivePacketsRoutine) (ha->pReturnReceivePacketsContext,
                                         pBufferCB);

    /* Keep track of RISC buffer pointer (for IP reinit) */
    ha->wIpBufferOut += wBufferCount;
    if (ha->wIpBufferOut >= IP_BUFFER_QUEUE_DEPTH)
        ha->wIpBufferOut -= IP_BUFFER_QUEUE_DEPTH;
} /* qla2x00_ip_receive_fastpost */


